
class LimitedIntSet{

      var limit : int;
      var arr : array<int>;
      var size : int;
     
     predicate Valid()
     reads this, this.arr;
     {
       arr != null &&
       0 <= size <= limit &&
       limit == arr.Length &&
       forall i,j :: 0 <= i < j < size ==> arr[i] != arr[j]
     }

      predicate Sorted()
      reads this, this.arr;
      requires Valid();
      {forall i, j :: 0 <= i < j < size ==> arr[i] <= arr[j]}


     method Init(lim : int)
     modifies this;
     requires lim > 0;
     ensures Valid();
     ensures size == 0 && limit == lim;
     ensures fresh(arr); // This is important, initialised with new array. Might otherwise cause strange compiler errors.
     {   
      limit := lim;
      arr := new int[lim];
      size := 0;   
     }

    // Returns true if elem is in the set
    function method contains (elem : int) :bool
    reads this, this.arr;
    requires Valid();
    {
     exists i :: 0 <= i < size && arr[i] == elem
    }
     
    //Adds elem to the set if it is not already present, or the set is full
    method Add(elem : int) returns (res :bool)
    modifies this.arr, `size;
    requires Valid();
    ensures Valid();
    
    ensures !old(contains(elem)) && old(size) < limit ==>
      res && contains(elem) && size == old(size+1) &&    
      (forall e :: e !=elem && old(contains(e)) ==> contains(e));
      
    ensures (old(contains(elem)) || old(size) >= limit) ==> 
        !res && size == old(size) && 
        forall i :: 0 <= i < size ==> arr[i] == old(arr[i]);           
        
    {if (!this.contains(elem) && size < limit)
       { 
         arr[size] := elem;
         size := size +1; 
         res := true;
       }
       else
        {
          res := false;
        }
     }
  
    
    // Returns the index where elem is present, or -1 if elem is not in the set.
    method Find(elem : int) returns (index : int)
    requires Valid();
    ensures 0 <= index ==> index < size && arr[index] == elem;
    ensures index < 0 ==> forall k :: 0 <= k < size ==> arr[k] != elem;  
    {
      index := 0;
        while (index < size)
         invariant 0 <= index <= size;
         invariant forall k :: 0 <= k < index ==> arr[k] != elem;
        {
          if(arr[index] == elem) {return;}
          index := index + 1;
        }
        index := -1;
    }

method BinarySearch(elem : int) returns (index : int)
  requires Valid() && Sorted();
  ensures 0 <= index ==> index < size && arr[index] == elem;
  ensures index < 0 ==> forall k :: 0 <= k < size ==> arr[k] != elem;  
  {
    var low := 0;
    var high := size;
    while(low < high)
    invariant 0<=low <= high <= size;
    invariant forall i :: 0 <= i < size && !(low <= i < high) ==> arr[i] != elem;
    {
      var mid := (low + high) / 2;
      if(arr[mid] < elem){
        low := mid + 1;
      } else if (elem < arr[mid]){
        high := mid;
      } else{
        index := mid;
        return;
      }
    }
    index := -1;
  }
    
    method main(){
      var s := new LimitedIntSet.Init(2);
      var b := s.Add(1);
      assert s.contains(1);
      var b2 := s.Add(1);
      assert !b2;
      var b3 := s.Add(2);
      assert s.contains(2);
      assert s.contains(1);
      var b4 := s.Add(3);
      assert !b4;
    }
    
}   

       


